/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "bpipep.h"
#include <bpipe/datatypes.h>

typedef struct
{
  size_t size;
  size_t offset;
} DataTypeMember;

typedef struct
{
  char *name;
  size_t size_struct;
  size_t size_raw;
  DataTypeMember *offsets;
  size_t n_offsets;
  void *init;
} DataType;

#define MEMBER DataTypeMember
#define STRUCT static DataType
#define STRUCT_ARRAY bpipe_datatypes

#include "datatype_info.h"

/*

  PROCEDURE
  bpipe_datatype_size - Determine the size (in bytes) of the
  internal storage requirements of a binary pipe data type.

  RETURNS

 It returns the size in bytes of the internal storage space required
  for an instance of the data.  This includes any structure padding.
  If the requested type doesn't exist, it returns @samp{0}.

*/
size_t
bpipe_datatype_size(
  BPDataType type	/* the data type */
)
{
  return ( type <=   _BPDType_NOTYPE || type >=  _BPDType_num ) ? 0 :
    bpipe_datatypes[type].size_struct;
}

/*

  PROCEDURE
  bpipe_datatype_raw_size - return the size (in bytes) of the
  raw storage requirements of a binary pipe data type.

  RETURNS
  It returns the size in bytes of the raw storage space required for
  an instance of the data.  It does not include any structure padding;
  just the total space directly used by the structure elements.  If
  the requested type doesn't exist, it returns @samp{0}.

*/
size_t
bpipe_datatype_raw_size(
  BPDataType type	/* the data type */
)
{
  return ( type <=   _BPDType_NOTYPE || type >=  _BPDType_num ) ? 0 :
    bpipe_datatypes[type].size_raw;
}

/*
  PROCEDURE
  bpipe_datatype_name - Attach a name to a @code{BPDataType}.

  RETURNS
  This routine will return a pointer to a string which contains the
  name of the datatype corresponding to the passed @code{BPDataType}
  value.  If the passed value is bogus, it returns @code{NULL}.
  Please note that the pointer returned points to the original copy of
  the field's name.  @strong{Do not change this data!}.

*/
const char *
bpipe_datatype_name(
  BPDataType	type	/* the datatype */
)
{

  return ( type < BPDType_NOTYPE || type >= BPDType_num ) ? NULL :
    bpipe_datatypes[type].name;
}

/*
  PROCEDURE
  bpipe_datatype_resolve - Identify a datatype.

  DESCRIPTION
  This routine will return a @code{BPDataType} corresponding to the
  passed string, which should contain the name of a datatype.

  RETURNS
  It returns @code{BPDType_NOTYPE} upon failure, else the code for
  the datatype.

*/
BPDataType
bpipe_datatype_resolve(
  const char *	datatype	/* a string with the name of a datatype */
)
{
  DataType *dt;

  for ( dt = bpipe_datatypes + 1 ; dt->name ; dt++)
    if ( ! strcmp( dt->name, datatype ) )
      return (BPDataType) (dt - bpipe_datatypes) ;

  return BPDType_NOTYPE;
}

/*

  PROCEDURE
  bpipe_datatype_write - Write a data type specification to an output channel.

  DESCRIPTION
  This routine writes a data type specification to the indicated
  output channel which is parseable by @code{proc_defs}.

  RETURNS
  It returns zero upon success, non-zero upon failure.

*/
int
bpipe_datatype_write(
  IOchannel *channel,		/* the output channel to which to write the
				   specification */
  BPDataType type		/* the data type */
)
{
  return
    bpipe_iochannel_write(bpipe_datatypes[type].name, (size_t) 1,
			  strlen(bpipe_datatypes[type].name), channel)
    !=
    strlen( bpipe_datatypes[type].name );
}

/*
   PROCEDURE
   bpipe_datatype_copy - call a copy function with offsets for a data
   type's atomic elements

   DESCRIPTION
   @code{bpipe_datatype_copy} issues copy commands to copy the
   elements in a data type without inter-element padding.  A routine
   must be provided with an interface similar to @code{memcpy} which
   will do the actual copying of data.  It assumes that all of the
   passed arguments are correct.

   RETURNS
   It returns zero upon success, non-zero if a copy operation returned
   failure.

*/
int
bpipe_datatype_copy(
  size_t	dst,		/* starting destination offset */
  size_t	src,		/* starting source offset */
  size_t	nobjs,		/* number of data objects to copy */

  BPDataType	type,		/* type of the data */
  BPXMap	map,		/* what type of mapping */
  void *	udata,		/* copy routine specific data */
  int		(*copy)(void *, size_t, size_t, size_t)
  				/* a function to copy the data */
)
{
  size_t n;

  switch ( map )
  {
  case BPXMap_Input_to_Core:
    /*
      dst and src are offsets in terms of the number
      of objects, not the number of bytes.  convert them to the latter
    */
    src *= bpipe_datatypes[type].size_raw;
    dst *= bpipe_datatypes[type].size_struct;

    /* if an object is an atomic (not conglomerate) type, and its
       raw size is the same as its struct size, we can do a
       straight copy of everything */
    if ( bpipe_datatypes[type].n_offsets == 1 &&
	 bpipe_datatypes[type].size_raw == bpipe_datatypes[type].size_struct )
    {
      if ( copy( udata, dst, src, nobjs * bpipe_datatypes[type].size_raw ) )
	return 1;
    }
    else    /* copy each object, element by element */
      for ( n = 0 ; n < nobjs ; n++, dst += bpipe_datatypes[type].size_struct )
      {
	size_t i;

	for ( i = 0 ; i < bpipe_datatypes[type].n_offsets ; i++ )
	{
	  size_t offset = bpipe_datatypes[type].offsets[i].offset;
	  size_t size   = bpipe_datatypes[type].offsets[i].size;

	  if ( copy( udata, dst + offset, src, size) )
	    return 1;
	  src += size;
	}
      }
    break;

  case BPXMap_Core_to_Core:
    /*
      dst and src are offsets in terms of the number
      of objects, not the number of bytes.  convert them to the latter
    */
    src *= bpipe_datatypes[type].size_struct;
    dst *= bpipe_datatypes[type].size_struct;

    /* copy the whole set of objects at once */
    if ( copy( udata, dst, src, nobjs * bpipe_datatypes[type].size_struct ) )
      return 1;
    break;

  case BPXMap_Core_to_Output:
    /*
      dst and src are offsets in terms of the number
      of objects, not the number of bytes.  convert them to the latter
    */
    src *= bpipe_datatypes[type].size_struct;
    dst *= bpipe_datatypes[type].size_raw;

    /* if an object is an atomic (not conglomerate) type, and its
       raw size is the same as its struct size, we can do a
       straight copy of everything */
    if ( bpipe_datatypes[type].n_offsets == 1 &&
	 bpipe_datatypes[type].size_raw == bpipe_datatypes[type].size_struct )
    {
      if ( copy( udata, dst, src, nobjs * bpipe_datatypes[type].size_raw ) )
	return 1;
    }
    else
      for ( n = 0 ; n < nobjs ; n++, src += bpipe_datatypes[type].size_struct)
      {
	size_t i;

	for ( i = 0 ; i < bpipe_datatypes[type].n_offsets ; i++ )
	{
	  size_t offset = bpipe_datatypes[type].offsets[i].offset;
	  size_t size   = bpipe_datatypes[type].offsets[i].size;

	  if ( copy( udata, dst, src + offset, size ) )
	    return 1;
	  dst += size;
	}
      }
    break;

  default:
    fprintf(stderr,
	    "bpipe: %s: %d: fatal internal error\n", __FILE__, __LINE__);
    exit( 1 );
  }

  return 0;
}

/*

  PROCEDURE
  bpipe_datatype_init - Fill memory with a datatype initialization structure.

  DESCRIPTION
  This routine will replicate a default datatype initialization
  structure throughout a region of memory.  The default initialization
  is the appropriate value of @samp{0}.  This routine is useful for
  intializaing a core image of a data packet field to a known state.

*/
void
bpipe_datatype_init(
  void *	dst,	/* the memory to init */
  BPDataType	type,	/* the type of the datatype */
  size_t	n	/* the number of instances of the datatype to init */
)
{
  bpipe_memfill( dst, n,
		 bpipe_datatypes[type].init, (size_t) 1,
		 bpipe_datatypes[type].size_struct );
}


/*
   PROCEDURE
   datatype_memcpy - @code{memcpy} type process function for
   @code{bpipe_datacopy_init_output}

   DESCRIPTION
   This routine is a process routine for @code{bpipe_datatype_copy}.  It
   does an immediate copy of the data chunk.  The @code{udata} passed
   to it must be a character array with two elements.  The first
   should point to the beginning of the destination data, the
   second should point to the beginning of the source data.

*/
static int
datatype_memcpy(void *udata, size_t dest, size_t src, size_t nbytes)
{
  memcpy(((char **) udata)[0] + dest, ((char **) udata)[1] + src, nbytes);
  return 0;
}

/*

  PROCEDURE
  bpipe_datatype_init_output - Fill memory with a datatype
  initialization structure.

  DESCRIPTION
  This routine will replicate a default datatype initialization
  structure for an output image of a field.  The default initialization
  is the appropriate value of @samp{0}.  It is somewhat inefficient,
  as the compacted version of the initialization structures should
  be created at compile time.

*/
void
bpipe_datatype_init_output(
  void *	dst,	/* the memory to init */
  BPDataType	type,	/* the type of the datatype */
  size_t	n	/* the number of instances of the datatype to init */
)
{
  char *start[2];

  if ( n == 0 )
    return;

  /*
     generate first initialization carefully, taking into account
     possible compaction of field in output image.
   */
  start[0] = dst;
  start[1] = bpipe_datatypes[type].init;

  bpipe_datatype_copy( (size_t) 0, (size_t) 0, (size_t) 1, type,
		       BPXMap_Core_to_Output, start, datatype_memcpy );

  /* now fill the rest of the memory with the compacted version */
  bpipe_memfill( (char *) dst + bpipe_datatypes[type].size_raw, n - 1,
		 dst, (size_t) 1, bpipe_datatypes[type].size_raw );
}

/*

  PROCEDURE
  bpipe_data_copy - copy data from one location to another

  DESCRIPTION
  @code{bpipe_data_copy} copies data of a given type and matrix
  specification.

  RETURNS
  nothing.

*/
void
bpipe_data_copy(
  void *dst,		/* the destination */
  void *src,		/* A pointer to the data */
  BPMatrix *matrix,	/* the data's matrix specification */
  BPDataType type	/* the data type */
)
{
  /* simple copy.  avoid using bpipe_datatype_size() for speed */
  memcpy(dst, src, matrix->ne * bpipe_datatypes[type].size_struct );
}

