/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <assert.h>

#include "bpipep.h"

/* ------------------------------------------------------------------ */

static int read_hdr( BPipe *bpipe );
static int read_hdr_defs(BPipe *bpipe, char **buf, size_t *buf_len);
static int read_dpkt_defs(BPipe *bpipe, char **buf, size_t *buf_len);
static int hdrf_channel_read(void *udata, size_t src, size_t dst,
			     size_t nbytes);

/* ------------------------------------------------------------------ */

static IOchannel *	hdrf_channel_read_ioc;

/* ------------------------------------------------------------------ */


/*

  PROCEDURE
  bpipe_input - attach an input stream to a binary pipe

  DESCRIPTION
  @code{bpipe_input} attaches an input stream to a binary pipe, opens it
  and reads the data stream's header.

  The parameter @code{input} is a string describing the input file or
  device associated with the input channel.  If it is the string
  @samp{stdin}, the input channel is associated with UNIX standard
  input.

  RETURNS
  It returns zero upon success, non-zero upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
      the passed path could not be opened.  Check @code{errno} for more
      information.

  @item BPEBADPIPE
    the header data definitions had errors

  @item BPEIOERROR
     an error occurred whilst reading the pipe

  @item BPENOMEM
      a memory allocation failed
  @end table

*/
int
bpipe_input(
  BPipe *bpipe,		/* a pointer to the binary pipe structure */
  const char *path 	/* a pointer to a string describing the input file or
  			   device to associate with the input channel */
)
{
  /* create and open input channel, if specified */

  if ( path == NULL || path[0] == '\0' )
  {
    bpipe_errno = BPEBADARG;
    errno = ENOENT;
    return 1;
  }

  if ( NULL == (bpipe->input = bpipe_iochannel_new( path )) )
    return 1;

  if ( bpipe_iochannel_open(bpipe->input, "r") )
  {
    bpipe_errno = BPENOFILE;
    return 1;
  }

  return read_hdr( bpipe );
}


/*

  PROCEDURE
  read_hdr - Read the header section of a binary pipe data stream.

  DESCRIPTION
  This routine reads and parses the header information of a binary
  pipe input data stream.  The header consists of header parameter and
  data packets field definitions and header parameter data.  The pipe
  I/O channel must already have been opened.

  RETURNS
  It returns zero upon success, non-zero otherwise.  Upon error
  @code{bpipe_errno} is set.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADPIPE
  the header data definitions had errors

  @item BPEIOERROR
  an error occurred whilst reading the pipe

  @item BPENOMEM
      a memory allocation failed
  @end table

*/
static int
read_hdr(
  BPipe *bpipe		/* the binary pipe from which to read */
)
{
  SLLNode *node;
  char *buf = NULL;
  size_t buf_len = 0;

  /* first read definitions of header parameters */
  if ( read_hdr_defs( bpipe, &buf, &buf_len ) )
    return 1;

  /* do the same for the data packet fields */
  if ( read_dpkt_defs( bpipe, &buf, &buf_len ) )
    return 1;

  free( buf );

  /*
     slurp down header fields and stick them in the appropriate place.
     this could be done with a linklist traverse call, but there's enough
     indirection involved, so it'll be done the brute force method here
   */

  /* tell copy routine where the input is */
  hdrf_channel_read_ioc = bpipe->input;

  for ( node = sll_head_node( bpipe->hdrf_ll ) ;
        node ;
        node = sll_next_node( bpipe->hdrf_ll, node ) )
  {
    HdrField *field = (HdrField *) sll_node_get_data( node );

    /*
       only process if it's not a new header field.  the user could have
       installed new fields before calling read_hdr
    */
    if ( field->new )
      continue;

    if ( bpipe_datatype_copy( (size_t) 0, (size_t) 0, field->matrix->ne,
			      field->type, BPXMap_Input_to_Core,
			      field->data, hdrf_channel_read ) )
    {
      bpipe_errno = BPEIOERR;
      return 1;
    }

  }

  return 0;

}


/*

  PROCEDURE
  read_hdr_defs - Read and store header parameter definitions.

  DESCRIPTION
  This routine reads header parameter definitions from a binary pipe,
  parses them, and stores the results in the binary pipe structure.

  RETURNS
  It returns zero upon success, non-zero upon failure.  It sets
  @code{bpipe_errno} upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code

  @item BPEBADPIPE
  the header data definitions had errors

  @item BPENOMEM
      a memory allocation failed
  @end table
*/
static int
read_hdr_defs(
  BPipe *	bpipe,		/* the binary pipe to read */
  char  **	buf,		/* the input buffer to use */
  size_t *	buf_len		/* the length of the input buffer */
)
{
  char *buffer;
  BPerrno bpipe_errno_s = bpipe_errno;

  int found_break = 0;

  bpipe_errno = BPNOERROR;

  bpipe->hdr_read = 1;

  while (( buffer = bpipe_iochannel_fgetrn( buf, buf_len, bpipe->input ) ))
  {
    char *name, *tmp;
    BPMatrix *matrix;
    BPDataType type;

    /* remove leading and trailing white space */
    while ( isspace( *buffer ) )
      buffer++;

    if ( buffer[0] == '\0' )	/* blank line is end of header defs */
    {
      found_break = 1;
      break;
    }

    tmp = buffer + strlen(buffer);
    while ( --tmp > buffer && isspace( *tmp ) )
      ;
    tmp[1] = '\0';

    if ( ( NULL == ( matrix = bpipe_proc_def( buffer, &name, &type ) ) ) ||
         bpipe_hdrf_add( bpipe, name, type, matrix, NULL, 1 )
        )
    {
      bpipe->hdr_read = 0;
      return 1;
    }

    bpipe->hdr_dpkts++;
  }

  bpipe->hdr_read = 0;

  /* if the input pipe is totally empty (i.e. if a previous program
     aborted, then we won't have found a break line between the header
     and data packet defs.  That's a pipe error. */
  if ( ! found_break )
    bpipe_errno = BPEBADPIPE;

  if ( bpipe_errno != BPNOERROR )
    return 1;

  bpipe_errno = bpipe_errno_s;
  return 0;
}


/*

  PROCEDURE
    read_dpkt_defs - Read data packet field definitions from a binary
    pipe.

  DESCRIPTION
    This routine reads in data packet field definitions from a binary
    pipe, parses them, and stores th results in the binary pipe
    structure.  It requires a line count variable, which it increments
    and uses for error output.  It uses @code{bpipe_proc_def} to parse
    the input.  an empty line is taken to signal the end of the data
    packet field definition section.

  RETURNS
    It returns zero upon success, non-zero upon failure.  It sets
    @code{bpipe_errno} upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code

  @item BPEBADPIPE
  the header data definitions had errors

  @item BPENOMEM
      a memory allocation failed
  @end table

  WARNING
    The header parameter definitions must be read before this routine is
    called.

*/
static int
read_dpkt_defs(
  BPipe *	bpipe,		/* the binary pipe to read */
  char  **	buf,		/* the input buffer to use */
  size_t *	buf_len		/* the length of the input buffer */
)
{
  char *buffer;
  BPerrno bpipe_errno_s = bpipe_errno;

  bpipe_errno = BPNOERROR;

  bpipe->dpkt_in_size = 0;
  while (( buffer = bpipe_iochannel_fgetrn( buf, buf_len, bpipe->input ) ))
  {
    char *name, *tmp;
    BPMatrix *matrix;
    BPDataType type;

    /* remove leading and trailing white space */
    while ( isspace( *buffer ) )
      buffer++;

    if ( buffer[0] == '\0' )	/* blank line is end of header defs */
      break;

    tmp = buffer + strlen(buffer);
    while ( --tmp > buffer && isspace( *tmp ) )
      ;
    tmp[1] = '\0';

    if ( ( NULL == ( matrix = bpipe_proc_def( buffer, &name, &type ) ) ) ||
         bpipe_dpktf_newP( bpipe, name, type, matrix, bpipe->dpkt_in_size,
			  BPDSite_INPUT) )
      return 1;

    bpipe->dpkt_in_size += matrix->ne * bpipe_datatype_raw_size(type);
  }

  if ( bpipe_errno != BPNOERROR )
    return 1;

  bpipe_errno = bpipe_errno_s;
  return 0;
}



/*

  PROCEDURE
  bpipe_proc_def - Process a field definition line.

  DESCRIPTION
  This routine takes a header or data packet field
  definition and determines the name of the field, its type, and
  its storage requirements.  Definitions have the format

    <data type>[<white space>]<name>[<matrix definition>]

  Blank lines should @emph{not} be passed to this routine, as they
  will be flagged as errors.  All leading and trailing white space
  should be removed before calling @code{proc_def}.

  RETURNS
  It returns a pointer to a @code{BPMatrix} structure if a definition
  was available and parsed correctly, @code{NULL} upon error.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADPIPE
    the definition was bad in some manner

  @item BPENOMEM
      a memory allocation failed
  @end table

*/
BPMatrix *
bpipe_proc_def(
  char *buf,			/* the buffer which contains the definition */
  char **name,			/* the address of a pointer which will point
				   to the name of the variable */
  BPDataType *type		/* storage type of parameter */
)
{
  size_t i;
  char *end, *tok;
  BPMatrix *matrix = NULL;

  /* first token on line is datatype */
  if ( NULL == ( end = strpbrk( buf, " \t" ) ) )
  {
    bpipe_errno = BPEBADPIPE;
    return NULL;
  }

  tok = buf;
  buf = end + 1;
  *end = '\0';

  if ( BPDType_NOTYPE ==
       (*type = bpipe_datatype_resolve( tok ) ) )
  {
    bpipe_errno = BPEBADPIPE;
    return NULL;
  }

  /*
    the next token is the name of the field.  either an open bracket
    or the end of the description ends the token
    */
  end = strchr( buf, '[' );

  *name = buf;
  if ( end )
  {
    buf = end + 1;
    *end = '\0';
  }
  else
    buf = NULL;

  if ( bpipe_check_field_name( *name ) )
  {
    bpipe_errno = BPEBADPIPE;
    return NULL;
  }

  /* deal with non-scalars */
  if ( buf != NULL )
  {
    char *tmp;
    int ket = 1;
    size_t nd = 0;

    /* count number of dimensions */
    for( tmp = buf ; *tmp ; tmp++ )
    {
      if      ( *tmp == ']' )
      {
	ket = 0;
	nd++;
      }
      else if ( *tmp == '[' )
      {
	if ( ket )		/* looking for ']' */
	{
	  bpipe_errno = BPEBADPIPE;
	  return NULL;
	}
	ket = 1;
      }
      else if ( ! isdigit(*tmp) )	/* check for bogus character */
      {
	bpipe_errno = BPEBADPIPE;
	return NULL;
      }
    }

    /* if ket is set, still looking for closing bracket */
    if ( ket )
    {
      bpipe_errno = BPEBADPIPE;
      return NULL;
    }

    matrix = bpipe_matrix_new( nd, NULL );

    /* things are simple from here on, as the dimension specification
       is guaranteed to be legal */

    tmp = strtok(buf, "]");
    for (matrix->ne = 1, i = 0 ; i < matrix->nd ; i++)
    {
      char *s;

      assert ( tmp != NULL );	/* oops */

      errno = 0;
      matrix->extent[i] = strtoul( tmp, &s, 10 );
      if ( *s || ERANGE == errno || 0 == matrix->extent[i] )
      {
	bpipe_errno = BPEBADPIPE;
	return NULL;
      }

      /* move past "][".  strtok collapses adjacent delimiters, so
         "][" is skipped completely */
      tmp = strtok(NULL, "][");

      matrix->ne *= matrix->extent[i];
    }
  }
  else	/* no dimensions specified, must be one, with one element */
    matrix = bpipe_matrix_new( (size_t) 1, NULL );

  return matrix;
}

/*

  PROCEDURE
  hdrf_channel_read - Read data from an input channel to memory.

  DESCRIPTION
  This is a callback routine for @code{bpipe_datatype_copy} which
  reads the requested number of bytes from the input channel in the
  file global variable @code{hdrf_channel_read_ioc} and deposits them
  in the memory pointed to by the @code{udata} parameter (with the
  appropriate offset as given by @code{bpipe_datatype_copy}.  It
  ignores the @code{src} argument.

  RETURNS
  It returns zero upon success, non-zero upon failure

*/

static int
hdrf_channel_read( void *udata, size_t src, size_t dst, size_t nbytes )
{
  return nbytes !=
    bpipe_iochannel_read( (char *) udata + dst, (size_t) 1, nbytes,
			  hdrf_channel_read_ioc );
}
