/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <linklist/linklist.h>

#include "bpipep.h"


/* ----------------------------------------------------------------- */

typedef struct
{
  SLinkList dpktf_ll;
  size_t size;
  size_t core_offset;
  size_t n_input_images;
} CreateMapData;

/* ----------------------------------------------------------------- */

/*
   this is a list of xmaps created by a callback routine called from
   bpipe_datatype_copy
*/
static SLinkList *xmap_ll;

/* ----------------------------------------------------------------- */

static Xmap *xmap_compact(size_t *nmap);
static int fill_dpktf_ll(void *data, void *udata);
static int fill_input_to_core_map(void *data, void *udata);
static int create_output_map(void *data, void *udata);
static int fill_core_to_output_map(void *data, void *udata);
static int xmap_process(void *udata, size_t dst, size_t src, size_t size);

/* ----------------------------------------------------------------- */

/*

  PROCEDURE
  bpipe_map - create transformation maps for data packet fields

  DESCRIPTION
  @code{bpipe_map} creates transformation maps between the input,
  core, and output images of a data packet.  It should be called after
  all manipulations of data packet field definitions are complete.  If
  the binary pipe has output channels, it must be called before
  writing the header information to the outputs, and should not be
  called afterwards.  It must be called prior to any data packet input
  or output activity on the binary pipe, and should not be called
  after any such activity.

  This routine should only be called once.  It will not re-map an
  already mapped @code{Bpipe}.

  RETURNS
  It returns the size of the core image in bytes, which should be used
  by the caller to allocate space for the data packets.  Upon error it
  returns @samp{0} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the @code{BPipe} has already been mapped
  @item a data packet field's input image matrix was bogus
  @item the mapping between a field's input and core images or
          output and core images was bogus
  @item the @code{BPipe} has an input stream and data packets
          are defined on the input stream, yet all of the
          data packet fields have been deleted
  @end itemize

  @item BPEBADPIPE
     either all of the input data packet fields were deleted or
     all of an output channel's data packet fields were deleted.

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
size_t
bpipe_map(
  BPipe *bpipe		/* a pointer to a binary pipe structure */
)
{
  CreateMapData cmd;

  cmd.size = bpipe->dpkt_in_size;

  if ( bpipe->hdr_mapped )
  {
    bpipe_errno = BPEBADARG;
    return 0;
  }

  bpipe->hdr_mapped = 1;
  bpipe->dpkt_core_size = 0;

  assert( NULL == bpipe->xmap );

  if ( rbtree_count(bpipe->dpktf) )
  {
    SLinkList dpktf_ll;

    /* allocate vector of pointers to the field descriptors */
    if ( NULL == ( dpktf_ll = sll_new( bpipe_dpktf_size_cmp ) ) )
    {
      bpipe_errno = BPENOMEM;
      return 0;
    }

    /*
      traverse the data packet field tree, creating a list of fields, sorted
      by the size of the fields' datatypes
    */
    if ( rbtree_utraverse( bpipe->dpktf, fill_dpktf_ll, dpktf_ll,
			   LEFT_TO_RIGHT ) )
      return 0;

    /* allocate the linked list containing the data packet field xmaps */
    if ( NULL == (xmap_ll = sll_new( NULL )) )
    {
      bpipe_errno = BPENOMEM;
      return 0;
    }

    /*
      traverse the data packets, generating the copy parameters required to
      copy from the input to the core images.  the parameters are stored
      in the file global linked list xmap_ll.  additionally, this pass
      determines the offsets in the core image of the fields.  it also
      keeps track of the number of input images (if any).
    */
    cmd.core_offset = 0;
    cmd.n_input_images = 0;
    if ( sll_utraverse( dpktf_ll, fill_input_to_core_map, &cmd ) )
      return 0;

    /* the core image size is the value of the core_offset after the last
       field is processed. */
    bpipe->dpkt_core_size = cmd.core_offset;

    /*
      compact maps, but only if there are some.  xmap_ll is destroyed by
      this operation
    */
    if ( sll_count( xmap_ll) )
      bpipe->xmap = xmap_compact( &bpipe->nmap );
    else if ( cmd.n_input_images != 0 )	/* all inputs deleted??? */
    {
      bpipe_errno = BPEBADPIPE;
      return 0;
    }
    else
      sll_delete( xmap_ll, free );

    cmd.dpktf_ll = dpktf_ll;

    /* loop through output channels and create output maps  */
    if ( sll_utraverse( bpipe->output, create_output_map, &cmd ) )
      return 0;

    sll_delete( dpktf_ll, SLL_NULL_DELETE );

  }

  /* allocate input buffer */
  if ( cmd.n_input_images )
  {
    bpipe->dpkt_inbuf_len = cmd.size;
    if ( NULL == ( bpipe->dpkt_inbuf = (char *) malloc( cmd.size ) ) )
    {
      bpipe_errno = BPENOMEM;
      return 0;
    }
  }

  /* check for weird error: input pipe, input data packet fields,
     but no maps, implying no input-to-core transfers?? */
  if ( bpipe->input && bpipe->nmap == 0 && bpipe->hdr_dpkts )
  {
    bpipe_errno = BPEBADARG;
    return 0;
  }

  return bpipe->dpkt_core_size;
}

/*

  PROCEDURE
  bpipe_map_alloc - create transformation maps for data packet fields
                    and allocate data packets

  DESCRIPTION
  @code{bpipe_map_alloc} is a convenience routine that calls
  @code{bpipe_map} to map the data packet, then allocates space for
  the requested number of data packets.  The space allocated is
  managed by @code{BPipe}; the calling program should @emph{not} free
  it.  If the parameter @code{size_p} is not @samp{NULL}, it should
  point to an integer where the size of the data packet will be
  stored.

  This routine should only be called once.  It will not re-map an
  already mapped @code{BPipe}.

  RETURNS
  It returns a pointer to the allocated space upon sucess.
  Upon error it returns @samp{NULL} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the number of data packets requested was non-positive
  @item the @code{BPipe} has already been mapped
  @item a data packet field's input image matrix was bogus
  @item the mapping between a field's input and core images or
          output and core images was bogus
  @item the @code{BPipe} has an input stream and data packets
          are defined on the input stream, yet all of the
          data packet fields have been deleted
  @end itemize

  @item BPEBADPIPE
     either all of the input data packet fields were deleted or
     all of an output channel's data packet fields were deleted.

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
void *
bpipe_map_alloc(
  BPipe *	bpipe,		/* a pointer to a binary pipe structure */
  int		ndpkts,		/* the number of data packets to allocate */
  size_t *	size_p		/* if not @samp{NULL}, a location where the
				   size of the data packet will be stored */
)
{
  size_t size;

  if ( ndpkts <= 0 )
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  size = bpipe_map( bpipe );

  if ( size_p )
    *size_p = size;

  if ( ! size )
    return NULL;

  assert( NULL == bpipe->dpkt );

  bpipe->dpkt = malloc( (size_t) ndpkts * size );
  if ( NULL == bpipe->dpkt )
    bpipe_errno = BPENOMEM;

  return bpipe->dpkt;
}



/*

  PROCEDURE
  xmap_compact - Compact a list of transformation maps

  DESCRIPTION
  @code{xmap_compact} operates on the file global linked list
  @code{xmap_ll}. This list consists of @code{Xmap} structures which
  specify the copy operations required to map between two data packet images.
  @code{xmap_compact} combines contigous copy operations.  @code{xmap_ll}
  @emph{must} contain at least one element.

  @code{xmap_ll} is deleted after it is processed, even upon error.

  RETURNS
  On error it returns @code{NULL} and sets @code{bpipe_errno}, else it returns
  a pointer to an array of @code{Xmap} structures which define the
  transformations.  It returns the length of the array via the
  parameter @code{nmap}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
    the list had no entries.

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
static Xmap *
xmap_compact(
  size_t *nmap		/* the number of elements in the resultant
			   @code{Xmap} array */
)
{
  Xmap *xmap, *xmap0;
  SLLNode *node, *next;
  Xmap *node_map;

  node = sll_head_node( xmap_ll );
  assert( node != NULL );
  node_map = (Xmap *) sll_node_get_data( node );

  while ( node &&  ( next = sll_next_node( xmap_ll, node ) ))
  {
    Xmap *next_map = (Xmap *) sll_node_get_data( next );

    if ( node_map->src + node_map->size == next_map->src &&
	 node_map->dst + node_map->size == next_map->dst )
    {
      node_map->size += next_map->size;
      sll_destroy_node( xmap_ll, next );
      free( next_map );
    }
    else
    {
      node = next;
      node_map = next_map;
    }
  }

  /* create array of maps */
  if ( NULL ==
       ( xmap0 = (Xmap *) malloc( sll_count( xmap_ll ) * sizeof(Xmap)) ) )
  {
    sll_delete( xmap_ll, free );
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  xmap = xmap0;

  node = sll_head_node ( xmap_ll );

  while ( node )
  {
    memcpy( xmap++, sll_node_get_data( node ), sizeof(Xmap) );
    node = sll_next_node( xmap_ll, node );
  }

  *nmap = sll_count( xmap_ll );
  sll_delete( xmap_ll, free );

  xmap_ll = NULL;

  return xmap0;
}

/*

  PROCEDURE
  fill_dpktf_ll - add the next link in a list

  DESCRIPTION
  This routine is an action function called when parsing the list of
  data packet fields.  It is passed a data field pointer and a list
  pointer (as user data) and inserts the field pointer into the list.

  RETURNS
  It returns zero upon success, non-zero if it couldn't insert the
  the data into the list.

*/
static int
fill_dpktf_ll(void *data, void *udata)
{
  SLinkList *ll = (SLinkList *) udata;

  if ( sll_insert(ll, data) )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  return 0;
}

/*

  PROCEDURE
  fill_input_to_core_map - Given a data packet field, determine the copy
  operations necessary to map from input to core state.

  DESCRIPTION
  This routine is an action routine passed to the linked list package.  It
  takes a passed data packet field and generates a list of copy operations
  which map the data packet field from its input state to its core state.
  It uses and updates a passed offset from the start of the core image.

  RETURNS
  It returns zero upon success, non-zero upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
    the data packet's input matrix was bogus

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
static int
fill_input_to_core_map( void *data, void *udata )
{
  DpktField *	dpktf = (DpktField *) data;
  CreateMapData *cmd = (CreateMapData *) udata;

  BPMatrix *core;

  /*
    core matrix spec may be NULL, indicating that it's the same as the
    input's.  don't need to duplicate the input's in that case. it'll
    never be the case that both core and input are NULL
  */
  if (dpktf->core)
  {
    if ( bpipe_matrix_verify(dpktf->core) )
    {
      bpipe_errno = BPEBADARG;
      return 1;
    }
    core = dpktf->core;
  }
  else
    core = dpktf->input;

  assert( core != NULL );

  /* verify matrix representation */
  if ( bpipe_matrix_verify( core ) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  /* set offset into output image */
  dpktf->core_offset = cmd->core_offset;

  /* increment offset for next field */
  cmd->core_offset += core->ne * bpipe_datatype_size( dpktf->type );

#if 0
  fprintf( stderr, "ne: %d size: %d core_offset: %d\n",
	   core->ne, bpipe_datatype_size( dpktf->type ),
	   cmd->core_offset );
#endif

  /* worry about mapping only if there's an input image */
  if (dpktf->input)
  {
    size_t start[2];
    cmd->n_input_images ++;

    /* verify matrix representation */
    if ( bpipe_matrix_verify(dpktf->input) )
    {
      bpipe_errno = BPEBADARG;
      return 1;
    }

    /* expand matrix map.  bpipe_errno is set upon error */
    if ( NULL ==
	 (dpktf->input_to_core = bpipe_matrix_map_expand(dpktf->input, core,
					   dpktf->input_to_core) ) )
      return 1;

    start[0] = dpktf->input_offset;
    start[1] = dpktf->core_offset;

    if ( bpipe_matrix_copy(dpktf->type, BPXMap_Input_to_Core,
			   dpktf->input, core,
			   dpktf->input_to_core, start,
			   xmap_process ) )
      return 1;

    bpipe_matrix_map_delete( dpktf->input_to_core );
    dpktf->input_to_core = NULL;	/* flag it as deleted */
  }

  return 0;
}

/*

  PROCEDURE
  create_output_map - Create an output map for an output channel.

  DESCRIPTION
  This routine creates a core image to output image map for a
  particular output channel.  It is an action routine called with the
  output channel as the passed data.

  RETURNS
  It returns zero upon success, non-zero upon error.  It sets
  @code{bpipe_errno} upon error.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADPIPE
    there were no output maps

  @item BPENOMEM
    a memory allocation failed

  @end table

 */
static int
create_output_map( void *data, void *udata )
{
  BPipeOutput *output = (BPipeOutput *) data;
  CreateMapData *cmd = (CreateMapData*) udata;
  Xmap *xmap;

  /* allocate the linked list containing the data packet field maps */
  if (NULL == (xmap_ll = sll_new( NULL )) )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  assert( output->xmap == NULL );

  if ( sll_utraverse( cmd->dpktf_ll, fill_core_to_output_map, output ) )
    return 1;

  if ( 0 == sll_count( xmap_ll ) )	/* no output maps?? */
  {
    bpipe_errno = BPEBADPIPE;
    return 1;
  }
  
  xmap = (Xmap *) sll_tail( xmap_ll );
  output->dpkt_size = xmap->dst + xmap->size;

  if ( NULL == ( output->xmap = xmap_compact(&output->nmap) ) )
      return 1;

  /* allocate an output transfer buffer if the mapping between
     the core and output images is not exact */
  if ( ! ( output->nmap == 1 && output->xmap->size == output->dpkt_size ) )
  {
    SLLNode node;
    if ( NULL == ( output->buf = (char *) malloc( output->dpkt_size ) ) )
    {
      bpipe_errno = BPENOMEM;
      return 1;
    }

    /* initialize the fields in the output map.  guards against
       writing uninitialized data if a field is resized from core to
       output */
    node = sll_head_node( output->dpktf );
    while ( node )
    {
      BPipeOutputDpktFInfo *info = sll_node_get_data(  node );

      bpipe_datatype_init_output( output->buf + info->output_offset,
			   info->field->type, info->n );
      node = sll_next_node( output->dpktf, node );
    }
  }

  return 0;
}

/*

  PROCEDURE
  fill_core_to_output_map - Given a data packet field, determine the copy
  operations necessary to map from core to output state.

  DESCRIPTION
  This routine is an action routine passed to the linked list package.  It
  takes a passed data packet field and generates a list of copy operations
  which map the data packet field from its core state to its output state.
  It uses and updates a passed offset from the start of the output image.

  RETURNS
  It returns zero upon success, non-zero upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
    the data packet's input matrix was bogus

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
static int
fill_core_to_output_map(void *data, void *udata)
{
  DpktField *	dpktf = (DpktField *) data;
  BPipeOutput *output = (BPipeOutput *) udata;

  BPipeOutputDpktFInfo *dpktf_info;
  Xmap *xmap;
  size_t offset[2];
  BPMatrix *output_matrix, *core_matrix;
  BPMatrixMap *map;
  BPMatrixOverride *over;

  /* check if globally deleted or if deleted in this channel */
  if ( dpktf->global_delete ||
       rbtree_search(output->dpktf_dels, dpktf, RBTREE_NULL_CMP) )
    return 0;

  /*
    get description for core matrix. no need to verify it, it was
    done for input to core mapping
    */

  core_matrix = dpktf->core ? dpktf->core : dpktf->input;

  /*
    check if there's a matrix override entry for this field.
    if there is, and it has a NULL matrix entry, that means that
    the core matrix should be used.  if there's no matrix override
    entry, use the default map from core to output
  */
  if (( over = (BPMatrixOverride *)
	sll_search(output->matrix, dpktf, bpipe_dpktf_matrix_override_cmp) ))
  {
    output_matrix = over->matrix ? over->matrix : core_matrix;
    map = over->map;
  }
  else
  {
    output_matrix = core_matrix;
    map = NULL;
  }

  assert( ! bpipe_matrix_verify(output_matrix) );

  /* insure that matrix map is consistent with core and output matrices */
  if ( bpipe_check_matrix_copy( core_matrix, output_matrix, map ) )
    return 1;

  /* expand matrix map */
  if ( NULL ==
       (map = bpipe_matrix_map_expand(core_matrix, output_matrix,
					 map)) )
    return 1;

  /* determine offset into images */
  offset[0] = dpktf->core_offset;
  offset[1] = ( NULL == (xmap = (Xmap *) sll_tail( xmap_ll ) ) ) ?
                       0 : xmap->dst + xmap->size;

  if ( bpipe_matrix_copy(dpktf->type, BPXMap_Core_to_Output,
			 core_matrix, output_matrix,
			 map, offset, xmap_process ) )
    return 1;

  bpipe_matrix_map_delete( map );
  if ( over )
    over->map = NULL;	/* flag it as no longer there */

  /* add this to the output channels "ok" list */
  if ( NULL ==
       ( dpktf_info = ( BPipeOutputDpktFInfo *)
	 malloc( sizeof(BPipeOutputDpktFInfo) ) ) )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }
  dpktf_info->field = dpktf;
  dpktf_info->output_offset = offset[1];
  dpktf_info->n = output_matrix->ne;
  sll_insert_tail( output->dpktf, dpktf_info );

  return 0;
}

/*

  PROCEDURE
  xmap_process

  DESCRIPTION
  This is a callback routine invoked by @code{bpipe_datatype_copy}.
  Instead of copying, it stores the source, destination, and length in
  a file global linked list, which will be compacted by
  @code{xmap_compact} after all data packet fields have been
  processed.

  RETURNS
  It returns zero upon success, non-zero upon failure (out of memoryb).
  It sets @code{bpipe_errno} to @code{BPENOMEM} upon failure.

*/
static int
xmap_process(
  void *udata,		/* copy routine specific data */
  size_t dst,		/* destination of chunk */
  size_t src,		/* source of chunk */
  size_t size		/* size of chunk in bytes */
)
{
  Xmap * xmap = (Xmap *) malloc( sizeof(Xmap) );

  if ( NULL == xmap )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  xmap->src = ((size_t *)udata)[0] + src;
  xmap->dst = ((size_t *)udata)[1] + dst;
  xmap->size = size;

  if ( sll_insert_tail( xmap_ll, xmap ) )
  {
    free( xmap );
    bpipe_errno = BPENOMEM;
    return 1;
  }

  return 0;
}
