/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdarg.h>

#include "bpipep.h"

/*

   PROCEDURE
   bpipe_matrix_delete - free a matrix specification structure and its
   accompanying data.

   DESCRIPTION
   If the passed pointer is not @code{NULL}, the memory it points to
   is freed.  If the dimensional extent field is not @code{NULL}, that
   memory is freed as well.

*/
void
bpipe_matrix_delete(
  BPMatrix *matrix	/* the matrix specification to be freed */
)
{
  if (matrix)
  {
    if (matrix->extent)
      free(matrix->extent);

    free(matrix);
  }
}

/*

   PROCEDURE
   bpipe_matrix_dup - duplicate a @code{BPipe} matrix structure

   DESCRIPTION
   This routine duplicates a binary pipe matrix and its accompanying
   data.  It does not perform special processing for @code{NULL} matrix
   pointers.

   RETURNS
   On successful completion, a pointer to a new matrix structure is returned.
   upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code
   @item BPEBADARG
     the passed argument is bogus or the matrix specification was illegal

   @item BPENOMEM
     a memory allocation failed
   @end table

*/
BPMatrix *
bpipe_matrix_dup(
  BPMatrix *src		/* the matrix to duplicate */
)
{
  BPMatrix *matrix = NULL;

  if (src)
  {
    size_t *extent = bpipe_data_dup( src->extent, src->nd, sizeof(size_t), 0 );
    if ( extent != NULL )
      matrix = bpipe_matrix_new(src->nd, extent);
  }
  else
    bpipe_errno = BPEBADARG;

  return matrix;
}

/*

   PROCEDURE
   bpipe_matrix_new_va - create a binary pipe matrix specification

   DESCRIPTION
   This routine creates a binary pipe matrix specification of a
   given dimensionality.  It allocates memory for the matrix structure and
   its associated data.  To ease filling in the associated extents array,
   the extents are passed as a variable argument list.

   RETURNS
   Upon successful completion a pointer to the new matrix is returned.
   Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code
   @item BPEBADARG
     the passed argument is bogus or the matrix specification was illegal

   @item BPENOMEM
     a memory allocation failed
   @end table

*/
BPMatrix *
bpipe_matrix_new_va(
  size_t nd,		/* the number of dimensions in the matrix */
  ...			/* the extents of the dimensions. there must be
                           as many extents as there are dimensions */
)
{
  va_list ap;
  size_t i;
  BPMatrix *matrix;

  if (nd < 1)
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (matrix = (BPMatrix *) malloc(sizeof(BPMatrix)) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  matrix->nd = nd;
  matrix->ne = 1;

  if ( NULL ==
       (matrix->extent = (size_t *) malloc(matrix->nd * sizeof(size_t)) ) )
  {
    free(matrix);
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  va_start(ap, nd);

  for (i = 0 ; i < nd ; i++)
  {
    if ( (matrix->extent[i] = va_arg(ap, size_t)) < 1)
    {
      bpipe_errno = BPEBADARG;
      bpipe_matrix_delete(matrix);
      matrix = NULL;
      break;		/* leave for loop */
    }

    matrix->ne *= matrix->extent[i];
  }

  va_end(ap);

  return matrix;
}


/*

   PROCEDURE
   bpipe_matrix_new - create a binary pipe matrix specification

   DESCRIPTION
   This routine creates a binary pipe matrix specification of a given
   dimensionality.  It allocates memory for the matrix structure and
   its associated data.  If a premade extents array is available, the
   parameter @code{extent} should point to it, and the routine will
   use it as is, directing the @code{BPipe} cleanup code to take
   responsibility for freeing it.  The user should @emph{not} free it,
   nor should it pass it to any other @code{BPipe} routine which takes
   responsibility for freeing it.  If @code{extent} is @code{NULL},
   the extents array is initialized to all ones.

   RETURNS
   Upon successful completion a pointer to the new matrix is returned.
   Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code
   @item BPEBADARG
     the passed argument is bogus or the matrix specification was illegal

   @item BPENOMEM
     a memory allocation failed
   @end table

*/

BPMatrix *
bpipe_matrix_new(
  size_t nd,		/* the number of dimensions in the matrix */
  size_t extent[]	/* if non @code{NULL}, points to an extents array to be
  			   duplicated */
)
{
  size_t i;
  BPMatrix *matrix;

  if (nd < 1)
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (matrix = malloc(sizeof(BPMatrix))) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  matrix->nd = nd;
  matrix->ne = 1;

  if ( extent )
  {
    matrix->extent = extent;
    for ( i = 0 ; i < nd ; i++)
      matrix->ne *= matrix->extent[i];
  }
  else
  {
    if ( NULL == (matrix->extent = malloc(matrix->nd * sizeof(size_t))) )
    {
      free(matrix);
      bpipe_errno = BPENOMEM;
      return NULL;
    }
    for ( i = 0 ; i < nd ; i++)
      matrix->extent[i] = 1;
  }

  return matrix;
}


/*

   PROCEDURE
   bpipe_matrix_verify - determine if a matrix specification is valid.

   DESCRIPTION
   The number of dimensions and the extent of the dimensions of a
   matrix specification are checked to determine if they are positive.
   the number of elements in the matrix are recalculated.  If the
   passed pointer is @code{NULL}, the routine simply returns.

   RETURNS
   If the specification is valid it returns zero, else non-zero.

*/
int
bpipe_matrix_verify(
  BPMatrix *matrix	/* the matrix specification to validate. */
)
{
  size_t i;
  
  if (matrix)
  {
    if (matrix->nd < 1)
      return 1;

    matrix->ne = 1;
    for (i = 0 ; i < matrix->nd ; i++)
    {
      if (matrix->extent[i] < 1)
	return 1;
      matrix->ne *= matrix->extent[i];
    }
  }

  return 0;
}


#ifndef BPIPEMINVAL
#define BPIPEMINVAL( a, b ) ((a) < (b) ? (a) : (b))
#endif

#ifndef BPIPEMAXVAL
#define BPIPEMAXVAL( a, b ) ((a) < (b) ? (b) : (a))
#endif

/*
   PROCEDURE
   bpipe_matrix_min - create a binary pipe matrix with the minimum
   number of dimensions and extent from a pair of binary pipe matrix.

   DESCRIPTION
   This routine creates a binary pipe matrix which has the minimum
   number of dimensions and extent from a pair of binary pipe matrix
   argument. It allocates memory for the matrix structure and its 
   associated data.

   RETURNS
   Upon successful completion a pointer to the new matrix is returned.
   Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code
   @item BPEBADARG
     the passed argument is bogus or the matrix specification was illegal

   @item BPENOMEM
     a memory allocation failed
   @end table

   */

BPMatrix *bpipe_matrix_min( BPMatrix *a, BPMatrix *b ) {

  size_t i;
  size_t nd;
  size_t *extent = NULL;
  BPMatrix *matrix = NULL;

  nd = BPIPEMINVAL( a->nd, b->nd );

  extent = (size_t *) malloc( nd * sizeof( size_t ) );
  if ( NULL == extent ) {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  for ( i = 0; i < nd; i++ )
    extent[ i ] = BPIPEMINVAL( a->extent[ i ], b->extent[ i ] );
  
  matrix = bpipe_matrix_new( nd, extent );

  return matrix;

}


/*
   PROCEDURE
   bpipe_matrix_max - create a binary pipe matrix with the maximum
   number of dimensions and extent from a pair of binary pipe matrix.

   DESCRIPTION
   This routine creates a binary pipe matrix which has the maximum
   number of dimensions and extent from a pair of binary pipe matrix 
   arguments. It allocates memory for the matrix structure and its 
   associated data.

   RETURNS
   Upon successful completion a pointer to the new matrix is returned.
   Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code
   @item BPEBADARG
     the passed argument is bogus or the matrix specification was illegal

   @item BPENOMEM
     a memory allocation failed
   @end table

   */


BPMatrix *bpipe_matrix_max( BPMatrix *a, BPMatrix *b ) {

  size_t i;
  size_t nd;
  size_t *extent = NULL;
  BPMatrix *matrix = NULL;

  nd = BPIPEMAXVAL( a->nd, b->nd );

  extent = (size_t *) malloc( nd * sizeof( size_t ) );
  if ( NULL == extent ) {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  for ( i = 0; i < nd; i++ )
    extent[ i ] = BPIPEMAXVAL( a->extent[ i ], b->extent[ i ] );
  
  matrix = bpipe_matrix_new( nd, extent );

  return matrix;

}


/*
   PROCEDURE
   bpipe_matrix_squeeze - remove all high order singleton dimensions

   DESCRIPTION
   This routine returns a copy of the passed matrix with all high
   order dimensions of extent one removed.  For example,
   @samp{[1][2][3][1]} becomes @samp{[1][2][3]}.

   RETURNS
   Upon successful completion a pointer to the new matrix is returned.
   Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code

   @item BPENOMEM
     a memory allocation failed
   @end table

*/
BPMatrix *bpipe_matrix_squeeze( BPMatrix *matrix )
{
  int i;
  int nd;
  size_t *extent = NULL;

  nd = matrix->nd;
  for ( i = matrix->nd - 1 ; i >= 0 ; i-- )
  {
    if ( matrix->extent[i] == 1 ) nd--;
    else break;
  }

  if ( nd == 0 )
  {
    i = 0;
    nd = 1;
  }

  extent = (size_t *) malloc( nd * sizeof( size_t ) );
  if ( NULL == extent )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  for ( ; i >= 0 ; i-- )
    extent[i] = matrix->extent[i];
  
  return bpipe_matrix_new( nd, extent );
}

/*
   PROCEDURE
   bpipe_matrix_crunch - remove all singleton dimensions

   DESCRIPTION

   This routine returns a copy of the passed matrix with all
   dimensions of extent one removed.  For example, @samp{[1][2][1]}
   becomes @samp{[2]}.

   RETURNS
   Upon successful completion a pointer to the new matrix is returned.
   Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following errors:

   @table @code

   @item BPENOMEM
     a memory allocation failed
   @end table

*/
BPMatrix *bpipe_matrix_crunch( BPMatrix *matrix )
{
  int i;
  int j;
  int nd;
  size_t *extent = NULL;

  nd = matrix->nd;
  for ( i = matrix->nd - 1 ; i >= 0 ; i-- )
  {
    if ( matrix->extent[i] == 1 ) nd--;
  }

  if ( nd == 0 )
    nd = 1;

  extent = (size_t *) malloc( nd * sizeof( size_t ) );
  if ( NULL == extent )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }


  for ( i = j = 0 ; i < matrix->nd ; i++ )
  {
    if ( matrix->extent[i] == 1 )
      continue;
    extent[j++] = matrix->extent[i];
  }

  /* nothing is != 1, so must be a point */
  if ( j == 0 )
    extent[0] = 1;
  
  return bpipe_matrix_new( nd, extent );
}

