/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "bpipep.h"

/*

  PROCEDURE
  bpipe_matrix_copy - Copy all or part of an N-dimensional matrix to a
  destination matrix.

  DESCRIPTION
  This routine analyzes the structure of the matrix (or submatrix) to
  be copied and the destination matrix to ensure that the largest
  possible chunks of memory are transferred.  It passes information
  about contigous chunks to a user supplied function, which is
  responsible for the actual copying operation.  The user supplied function
  is called by @code{bpipe_datatype_copy}.

  RETURNS
  It returns zero upon success, non-zero otherwise.  Upon error
  @code{bpipe_errno} is set.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
      the matrix copy request was inconsistent with the sizes of the
      source and destination matrices or there were non-positive
      extents.

  @item BPEIOERR
      one of the copy operations failed

  @item BPENOMEM
      a memory allocation failed
  @end table

*/
int
bpipe_matrix_copy(
  BPDataType	type,	/* type of a datum */
  BPXMap	xmap,	/* what type of mapping */
  BPMatrix *	src,    /* description of the source matrix */
  BPMatrix *	dst,	/* description of the destination matrix */
  BPMatrixMap *	map,	/* the mapping from source to destination */

  void *	udata,	/* copy routine specific information */
  int (*copy)(void *udata, size_t dest, size_t src, size_t nbytes)
  	/* this will be called multiple times to copy the chunks of data */
)
{
  size_t nd, i, j;
  size_t nd_c, off;
  size_t srcp, dstp;
  size_t *extent;
  size_t *src_stride,	/* memory offsets between tensors of the same
			   dimensionality */
         *dst_stride,	/* memory offsets between tensors of the same
			   dimensionality */
         contig_cnt;	/* number of data in the largest contiguous
			   block of data */
  size_t *src_stride0, *dst_stride0;

  /* check that submatrix source and destination specifications are legal */
  if ( bpipe_check_matrix_copy(src, dst, map) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  extent = map->extent;

  /*
     determine dimensionality of the submatrix to be copied.
     for example, (3,1,1) is really 1D.
  */
  for (nd = src->nd ; nd > 1 && extent[nd-1] == 1 ; nd--)
    ;

  /*
     determine if the data in the lower dimensions are contiguous both at the
     source and the destination, and collapse as many dimensions as possible
     into one chunk.

     contig_cnt is the number of elements in the largest block of contiguous
     data.
   */
  contig_cnt = 1;

  /*
     this loop breaks at the first dimension for which the submatrix
     extent isn't the same as in the source or destination matrices.
     since all submatrices with one less dimension than this
     are contiguous, the adjacent submatrices in this level are
     contiguous.  this determines the maximum chunk of contigous
     data to copy.
  */
  for (nd_c = nd, i = 0 ; i < nd ; i++)
  {
    nd_c--;

    contig_cnt *= extent[i];

    /*
       check if this submatrix completely fills this dimension for
       both source and destination.  if so adjacent submatrices of one
       higher dimensionality are contiguous.
     */
    if ( ! (extent[i] == src->extent[i] && extent[i] == dst->extent[i]) )
      break;
  }

  /*
     after this point, the dimension of the data is described by the
     number of contiguous hunks of memory and thus the number of loops
     needed to copy it.  for example, 0 dimensions (a point) implies
     the data is one contiguous hunk of memory.

     since the smaller dimensions are collapsed, the offsets and
     extents that are important start at nd - nd_c - 1
  */
  off = nd - nd_c;
  nd = nd_c;

  src_stride0 = src_stride = malloc( src->nd * sizeof(size_t));
  dst_stride0 = dst_stride = malloc( dst->nd * sizeof(size_t));

  if (src_stride == NULL || dst_stride == NULL)
  {
    bpipe_errno = BPENOMEM;
    if ( src_stride ) free( src_stride0 );
    if ( dst_stride ) free( dst_stride0 );
    return 1;
  }

  /* generate data position increments for each dimension and the
     start of source data */
  srcp = map->src_off[0];
  src_stride[0] = 1;
  for (i = 1 ; i < src->nd ; i++)
  {
    src_stride[i] = src->extent[i - 1] * src_stride[i - 1];
    srcp += map->src_off[i] * src_stride[i];
  }

  /* generate data position increments for each dimension and the
     start of the destination data */
  dstp = map->dst_off[0];
  dst_stride[0] = 1;
  for (i = 1 ; i < dst->nd ; i++)
  {
    dst_stride[i] = dst->extent[i - 1] * dst_stride[i - 1];
    dstp += map->dst_off[i] * dst_stride[i];
  }

  /* advance arrays of importance to ease access */
  extent += off;
  src_stride += off;
  dst_stride += off;

  /* optimized routines for the most common cases */
  switch (nd)
  {
    case 0:      /* 1 chunk, real simple */
      if ( bpipe_datatype_copy(dstp, srcp, contig_cnt, type, xmap,
			       udata, copy) )
      {
        bpipe_errno = BPEIOERR;
        return 1;
      }
      break;

    case 1:      /* 1D */
      for (i = 0 ; i < extent[0] ; i++)
      {
        if ( bpipe_datatype_copy(dstp, srcp, contig_cnt, type, xmap, udata,
				 copy) )
        {
          bpipe_errno = BPEIOERR;
          return 1;
        }
        srcp += src_stride[0];
        dstp += dst_stride[0];
      }
      break;

    case 2:      /* 2D */
    {
      size_t srcp0 = srcp;
      size_t dstp0 = dstp;

      for (i = 0 ; i < extent[0] ; i++)
      {
        srcp = srcp0;
        dstp = dstp0;
        for (j = 0 ; j < extent[1] ; j++)
        {
	  if ( bpipe_datatype_copy(dstp, srcp, contig_cnt, type, xmap, udata,
				   copy) )
          {
            bpipe_errno = BPEIOERR;
            return 1;
          }
          srcp += src_stride[1];
          dstp += dst_stride[1];
        }
        srcp0 += src_stride[0];
        dstp0 += dst_stride[0];
      }
      break;
    }

    default:     /* nD */
    {
      /*
	 multidimensional copies are done in a brute force manner.
	 an index array keeps track of what the value of the array
	 indices for each dimension.  stride arrays provide increments
	 for moving to the next block of data when an index is
	 incremented.  the copying is done by copying the
	 smallest submatrix possible, updating indices, dealing with
	 reinitialization of sub-dimensional indices when an index
	 hits its limit and the next higher dimensional index is
	 incremented.
      */
      size_t *src_origin;         /* origin of the data at the first
                                     submatrix to be copied */
      size_t *dst_origin;         /* same, but for destination */
      size_t *idx;                /* index of data along a dimension
                                     relative to start */
      int  more_data;             /* flag indicating there's more data
				     to copy */
      size_t nchunks = extent[0];

      idx = (size_t *) malloc(nd * sizeof(size_t));
      src_origin = (size_t *) malloc(nd * sizeof(char *));
      dst_origin = (size_t *) malloc(nd * sizeof(char *));

      if ( NULL == idx || NULL == src_origin || NULL == dst_origin )
      {
	bpipe_errno = BPENOMEM;
	if (idx) free( idx );
	if (src_origin) free( src_origin );
	if (dst_origin) free( dst_origin );
	free( src_stride0 );
	free( dst_stride0 );
	return 1;
      }

      /*
         initialize dimensional index to 0 and setup source and destination
         start stacks.  the first elements of these arrays are never used.
      */
      for ( i = 1 ; i < nd ; i++ )
      {
        idx[i] = 0;
        src_origin[i] = srcp;
        dst_origin[i] = dstp;
        nchunks *= extent[i];
      }

      do
      {
        size_t dim;

        for (i = 0 ; i < extent[0] ; i++)
        {
	  if ( bpipe_datatype_copy(dstp, srcp, contig_cnt, type, xmap, udata,
				   copy) )
          {
            bpipe_errno = BPEIOERR;
            return 1;
          }
          srcp += src_stride[0];
          dstp += dst_stride[0];
        }

        /* done with this dimension, start moving to higher
           dimensions to find the next data to be copied

	   since there are at least 3 dimensions here,
	   it'd be possible to optimize this a bit further
	   (i.e., make the above loop a set of 3 nested loops
	   and change the stuff below to start work for
	   dim = 4).

	*/

        /* reset more_data flag. will be set if data is found */
        more_data = 0;

        for (dim = 1 ; dim < nd ; dim++)
          if (++idx[dim] < extent[dim])
          {
            /* there's more, so increment the start of the src and
               destination data stacks */
            srcp = src_origin[dim] += src_stride[dim];
            dstp = dst_origin[dim] += dst_stride[dim];

            /* set the stack pointers for the lower dimensions to
               point to the new starts */
            for (i = 1; i < dim ; i++)
            {
              src_origin[i] = srcp;
              dst_origin[i] = dstp;
              idx[i] = 0;
            }

            /* signal that more data has been found and all is ready */
            more_data++;
            break;
          }

      } while (more_data);

      free( idx );
      free( src_origin );
      free( dst_origin );
      break;
    }
  }

  free( src_stride0 );
  free( dst_stride0 );

  return 0;
}

/*

  PROCEDURE
  bpipe_check_matrix_copy - Ensure that a matrix copy request is legal.

  DESCRIPTION
  This routine checks that the dimensions of the submatrix are
  compatible with the destination matrix, that the dimensional extents
  are consistent with the source and destination matrices, and that
  the offsets are consistent.  It can deal with non-expanded
  map entries.

  RETURNS
  It returns zero if the request is legal, non-zero otherwise.

*/
int
bpipe_check_matrix_copy(
  BPMatrix *src,       	/* description of the source matrix */
  BPMatrix *dst,	/* description of the destination matrix */
  BPMatrixMap *map	/* description of source to destination mapping */
)
{
  size_t i, nd;

  /* if map is NULL, it's a default map, and there's nothing to do here */
  if ( NULL == map )
    return 0;

  /* ensure legal extents */
  if ( map->extent )
  {
    for (i = 0 ; i < src->nd ; i++)
      if (map->extent[i] < 1)
	return 1;
    /* determine dimensionality of submatrix and ensure that it'll fit */
    for (nd = src->nd ; nd > 1 && map->extent[nd - 1] == 1 ; nd--)
      ;
  }
  else
    nd = dst->nd;

  if (nd > dst->nd)
    return 1;

  for (i = 0 ; i < nd ; i++ )
    if (
	( map->src_off &&
	    ( map->extent &&
	      map->src_off[i] + map->extent[i] > src->extent[i] )
	)
	||
	( map->dst_off &&
	    ( map->extent &&
	      map->dst_off[i] + map->extent[i] > dst->extent[i] )
	)
       )
      return 1;

  return 0;

}
