/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdarg.h>

#include "bpipep.h"

/*

  PROCEDURE
  bpipe_matrix_map_new - allocate a matrix map structure and its
  associated arrays.

  DESCRIPTION
  A matrix map structure is allocated and its associated arrays
  (source and destination offsets and copied submatrix extents) are
  constructed using the passed data (if available).  It does not
  automatically create the new map's associated arrays if they are
  @code{NULL}.

  Note that @code{bpipe_matrix_map_new} uses the offset and extent
  arrays as is, directing the @code{BPipe} cleanup code to take
  responsibility for freeing them.  The application should @emph{not}
  free them, nor should it pass them to any other @code{BPipe} routine
  which takes responsibility for freeing them.  After passing them to
  @code{bpipe_matrix_map_new}, it should not alter them in any
  fashion.

  RETURNS
  It returns a pointer to a newly create matrix map structure upon
  success.  Upon error @code{bpipe_errno} is set and @code{NULL} is
  returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
      the passed argument is bogus or the matrix specification was illegal

  @item BPENOMEM
      a memory allocation failed
  @end table

*/
BPMatrixMap *
bpipe_matrix_map_new(
  BPMatrix *src,	/* source matrix spec */
  size_t src_off[],	/* offset of source submatrix to copy to destination.
                           extent is equal to the dimensions of the source
                           matrix. */
  BPMatrix *dst,	/* destination matrix spec */
  size_t dst_off[],	/* offset of destination submatrix.  extent is equal
                           to the dimensions of the destination matrix. */
  size_t extent[]	/* the extents of the submatrix to copy.  the
  			   extent of this array is equal to the
  			   dimensions of the source matrix. */
)
{
  BPMatrixMap *map;

  if ( (src && src->nd < 1) || (dst && dst->nd < 1) )
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (map = malloc(sizeof(BPMatrixMap))) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  map->src_off = src_off;
  map->dst_off = dst_off;
  map->extent  = extent;

#if 0

  map->src_off = src_off ? src_off : bpipe_offset_new( src->nd, 0 );
  map->dst_off = dst_off ? dst_off : bpipe_offset_new( dst->nd, 0 );
  map->extent  = extent  ? extent  : bpipe_extent_new( src->nd, 1 );

  if ( NULL == map->src_off || NULL == map->dst_off || NULL == map->extent )
  {
    if ( map->src_off ) free( map->src_off );
    if ( map->dst_off ) free( map->dst_off );
    if ( map->extent  ) free( map->extent  );
    free( map );
    return NULL;
  }
#endif
  
  return map;
}


/*

  PROCEDURE
  bpipe_matrix_map_expand - expand all @code{NULL} matrix map entries to
  create default maps

  DESCRIPTION
  Matrix map structures may contain @code{NULL} entries indicating
  that default maps are to be performed.  Before creating copy maps
  these must be expanded.  This routine will, if necessary, fill in
  @code{NULL} matrix map entries.  In the case that there are
  @code{NULL} entries in the structure, a new matrix map structure
  will be created and all available data duplicated.  If offsets are
  not provided, they are set to the origin if the matrices.  If the
  copied matrix extents are not provided, the extents of the largest
  submatrix which can be copied are used.

  If the pointer to the source matrix map is @code{NULL}, it is
  treated as if all of the matrix map entries were @code{NULL}.  in
  the case that a new map is created, the old is destroyed.

  The resultant map is @emph{not} checked to see if it is legal.

  RETURNS
  It returns a pointer to the original matrix if nothing needed to be
  expanded, else a newly created matrix map structure.  The original
  map is destroyed if a new is created.  Upon error @code{bpipe_errno}
  is set and @code{NULL} is returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
    the passed source or destination pointers are bogus

  @item BPENOMEM
    a memory allocation failed
  @end table

*/
BPMatrixMap *
bpipe_matrix_map_expand(
  BPMatrix *src,	/* source matrix spec */
  BPMatrix *dst,	/* destination matrix spec */
  BPMatrixMap *map	/* the matrix map to expand */
)
{
  BPMatrixMap *tmap;

  if (src == NULL || dst == NULL)
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  /*
     if the map doesn't exist, create it. duplicate it if it does.
     a side effect of the duplication is the extension of NULL offset or
     extent arrays. these will be set to 0.
  */
  if (map)
    tmap = bpipe_matrix_map_dup( src, dst, map, 1 );
  else
  {
    map = bpipe_matrix_map_new( src, NULL, dst, NULL, NULL );
    tmap = bpipe_matrix_map_dup( src, dst, map, 1 );
  }

  if (!tmap)
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  /* if no extent was passed, find largest submatrix that will fit */
  if ( NULL == map->extent )
  {
    size_t i;
    size_t nd = src->nd > dst->nd ? dst->nd : src->nd;

    for (i = 0 ; i < nd ; i++ )
    {
      size_t d_src = src->extent[i] - (map->src_off ? map->src_off[i] : 0);
      size_t d_dst = dst->extent[i] - (map->dst_off ? map->dst_off[i] : 0);

      tmap->extent[i] = d_src > d_dst ? d_dst : d_src;
    }

    for (; i < src->nd ; i++ )
      tmap->extent[i] = 1;
  }

  bpipe_matrix_map_delete( map );

  return tmap;
}


/*

  PROCEDURE
    bpipe_matrix_map_dup - duplicate an existing matrix map structure

  DESCRIPTION
    this routine creates a new matrix map structure and duplicates all
    available data in the passed matrix map.  if the flag
    @code{create_empty} is non-zero, memory will be allocated and
    cleared for the @code{NULL} entries in the source matrix map.

  RETURNS
    it returns a pointer to the new matrix structure upon success.
    if the passed matrix map pointer is @code{NULL}, it returns @code{NULL}.
    upon error @code{bpipe_errno} is set and @code{NULL} is returned.

  ERRORS
    upon error @code{bpipe_errno} is set to one of the following errors:

    @table @code
    @item BPEBADARG
      the passed argument is bogus or the matrix specification was illegal

    @item BPENOMEM
      a memory allocation failed
    @end table
    @c	need this to foil c2man's punctuation fixups

*/

BPMatrixMap *
bpipe_matrix_map_dup(
  BPMatrix *src,	/* source matrix spec */
  BPMatrix *dst,	/* destination matrix spec */
  BPMatrixMap *src_map,	/* the matrix map to be duplicated */
  int create_empty	/* if non zero, allocate and clear memory
  			   for @code{NULL} map structure entries */
)
{
  BPMatrixMap *map = NULL;

  if (src_map)
  {
    /* save bpipe_errno */
    BPerrno bpipe_errno_s = bpipe_errno;
    bpipe_errno = BPNOERROR;
 
    if (! (map = malloc(sizeof(BPMatrixMap))) )
    {
      bpipe_errno = BPENOMEM;
      return NULL;
    }

    /*
       if any of the offset specs or the extent spec is NULL, bpipe_data_dup
       will return NULL without error, so can't use a NULL return to check
       for out of memory situations for later clean up.  use bpipe_errno.
     */
     
    map->src_off = bpipe_data_dup(src_map->src_off, src->nd, sizeof(size_t),
				  create_empty);
    map->dst_off = bpipe_data_dup(src_map->dst_off, dst->nd, sizeof(size_t),
				  create_empty);
    map->extent  = bpipe_data_dup(src_map->extent,  src->nd, sizeof(size_t),
				  create_empty);

    if (bpipe_errno != BPNOERROR)
    {
      if (map->src_off) free(map->src_off);
      if (map->dst_off) free(map->dst_off);
      if (map->extent)  free(map->extent);
      free(map);
      return NULL;
    }

    bpipe_errno = bpipe_errno_s;
  }

  return map;
}


/*

  PROCEDURE
    bpipe_matrix_map_delete -
    free a matrix map structure and its accompanying data.

  DESCRIPTION
    if the passed pointer is not @code{NULL}, the memory it points to
    is freed along with the memory pointed to be the non @code{NULL}
    fields in the structure.

*/
void
bpipe_matrix_map_delete(
  BPMatrixMap *map	/* the matrix map to delete */
)
{
  if ( map )
  {
    if ( map->src_off ) free( map->src_off );

    if ( map->dst_off ) free( map->dst_off );

    if ( map->extent  ) free( map->extent  );

    free( map );
  }
}
