/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "bpipep.h"

/*

  PROCEDURE
  bpipe_data_dup - allocate memory for and duplicate a contiguous list of
                   objects

  DESCRIPTION

  @code{bpipe_data_dup} allocates a block of memory and copies a
  sequential list of objects to that memory.  If the pointer to the
  original data is @code{NULL}, and the @code{create_empty} flag is
  non zero, a block of memory large enough to contain the specified
  data is created and cleared.  If the flag is zero, @code{NULL} is
  returned.

  RETURNS
  It returns a pointer to a block of memory duplicating the original
  data.  Upon error @code{bpipe_errno} is set and @code{NULL} is
  returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPENOMEM
    a memory allocation failed
  @end table

*/
void *
bpipe_data_dup(
  void *data,		/* the buffer containing the data */
  size_t n,		/* the number of elements of data */
  size_t size,		/* the size of an element */
  int create_empty	/* if true, allocate and clear memory if the
  			   parameter data is NULL */
)
{
  void *p = NULL;

  if (data)
  {
    if ( NULL == (p = malloc(n * size)) )
    {
      bpipe_errno = BPENOMEM;
      return NULL;
    }
    memcpy(p, data, n * size);
  }
  else if ( create_empty )
  {
    if (NULL == (p = calloc( n, size )) )
    {
      bpipe_errno = BPENOMEM;
      return NULL;
    }
  }

  return p;
}



/*

  PROCEDURE
  bpipe_extent_new_va - allocate an extent array

  DESCRIPTION
  Extent arrays are required by various @code{bpipe} support routines.
  This routine makes it easy to allocate one and fill it in.

  RETURNS
  It returns a pointer to an array of integers, filled in with
  the specified values.  Upon error @code{bpipe_errno} is set and
  @code{NULL} is returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
  the number of dimensions or one of the extents was less than 1
  @item BPENOMEM
  a memory allocation failed
  @end table

*/
void *
bpipe_extent_new_va(
  size_t nd,		/* the number of dimensions */
  ...			/* the integer values of the extents. there must be
			   @code{nd} values, and should be of type @code{int} */
)
{
  size_t *extent = NULL;
  size_t i;
  va_list ap;

  if (nd < 1)
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (extent = malloc(nd * sizeof(size_t))) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }
  
  va_start(ap, nd);

  for (i = 0 ; i < nd ; i++)
  {
    extent[i] = va_arg(ap, int);

    if (extent[i] < 1)
    {
      bpipe_errno = BPEBADARG;
      free(extent);
      extent = NULL;
      break;
    }
  }

  va_end(ap);

  return extent;
}

/*

  PROCEDURE
  bpipe_extent_new - Allocate an extent array.

  DESCRIPTION
  Extent arrays are required by various @code{bpipe} support routines.
  This routine creates one and sets all of its elements to the
  passed value.

  RETURNS
  It returns a pointer to an array of integers, filled in with
  the specified value.  Upon error @code{bpipe_errno} is set and
  @code{NULL} is returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
  the number of dimensions or the extent was less than 1
  @item BPENOMEM
  a memory allocation failed
  @end table

*/
void *
bpipe_extent_new(
  size_t nd,		/* the number of dimensions */
  size_t init		/* the integer value of the extents */
)
{
  size_t *extent = NULL;
  size_t i;

  if (nd < 1 || init < 1)
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (extent = malloc(nd * sizeof(size_t))) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }
  
  for (i = 0 ; i < nd ; i++)
    extent[i] = init;

  return extent;
}

/*

  PROCEDURE
  bpipe_offset_new_va - Allocate an offset array.

  DESCRIPTION
  Offset arrays are required by various @code{bpipe} support routines.
  This routine makes it easy to allocate one and fill it in.

  RETURNS
  It returns a pointer to an array of integers, filled in with
  the specified values.  Upon error @code{bpipe_errno} is set and
  @code{NULL} is returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
  the number of dimensions was less than 1 or one of the offsets was negative
  @item BPENOMEM
  a memory allocation failed
  @end table

*/
size_t *
bpipe_offset_new_va(
  size_t nd,		/* the number of dimensions */
  ...			/* the integer values of the extents. there must be
			   @code{nd} values, and should be of type @code{int} */
)
{
  size_t *offset = NULL;
  size_t i;
  va_list ap;

  if (nd < 1)
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (offset = malloc(nd * sizeof(size_t))) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }
  
  va_start(ap, nd);

  for (i = 0 ; i < nd ; i++)
  {
    int offset_tmp = va_arg(ap, int);

    if (offset_tmp < 0)
    {
      bpipe_errno = BPEBADARG;
      free(offset);
      offset = NULL;
      break;
    }

    offset[i] = offset_tmp;
  }

  va_end(ap);

  return offset;
}

/*

  PROCEDURE
  bpipe_offset_new - Allocate an offset array.

  DESCRIPTION
  Offset arrays are required by various @code{bpipe} support routines.
  This routine makes it easy to allocate one.  It creates an
  offset array with all offsets set to a passed value.

  RETURNS
  It returns a pointer to an array of integers, filled in with
  the specified value.  Upon error @code{bpipe_errno} is set and
  @code{NULL} is returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code
  @item BPEBADARG
  the number of dimensions was less than 1
  @item BPENOMEM
  a memory allocation failed
  @end table

*/
size_t *
bpipe_offset_new(
  size_t	nd,		/* the number of dimensions */
  size_t	init		/* the value to initialize the offsets to */
)
{
  size_t *offset = NULL;
  size_t i;

  if ( nd < 1 )
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( NULL == (offset = malloc(nd * sizeof(size_t))) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }
  
  for (i = 0 ; i < nd ; i++)
    offset[i] = init;

  return offset;
}

/*

  PROCEDURE
  bpipe_check_field_name - Check that a field name is legal.

  DESCRIPTION
  Field names must start with a character from the alphabet or an
  underscore and can contain only alphanumeric and underline
  characters.  Field names beginning with an underscore are
  reserved for use by @code{bpipe}.
  

  RETURNS
  It returns zero if the name is legal, non-zero if it isn't

*/
int
bpipe_check_field_name(
  const char *name	/* the name to check */
)
{
  char c;
  
  if ( NULL == name ||
       '\0' == name[0]
       || ! ( isalpha(name[0]) || '_' == name[0]) )
    return 1;

  while ( (c = *name++) )
    if ( ! (isalnum( c ) || c == '_' ) )
      return 1;

  return 0;
}

/*

  PROCEDURE
  bpipe_memfill - Fill a region of memory with data.

  DESCRIPTION
  This routine will replicate user supplied data throughout a region
  of memory.  It uses @code{memcpy}, and attempts to invoke it as
  few times as possible.

*/  
void
bpipe_memfill(
  void *	dstp,		/* the memory to fill */
  size_t	n_dst,		/* the number of data elements in the
				   destination */
  void *	srcp,		/* the data to replicate */
  size_t	n_src,		/* the number of data elements in the
				   src */
  size_t	size		/* the size of a data element */
)
{
  size_t copy_size;
  size_t dst_size = n_dst * size;
  size_t src_size = n_src * size;

  char *dst;
  if ( 0 == n_dst || 0 == n_src )
    return;

  /* copy first batch */
  copy_size = dst_size > src_size ? src_size : dst_size;
  memcpy( dstp, srcp, copy_size );

  dst = (char *) dstp + copy_size;
  dst_size -= copy_size;
  src_size  = copy_size;
  srcp = dstp;

  while( dst_size )
  {
    copy_size = dst_size > src_size ? src_size : dst_size;
    memcpy( dst, srcp, copy_size );

    dst      += copy_size;
    dst_size -= copy_size;
    src_size += copy_size;
  }
}
