/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <errno.h>

#include <rbtree/rbtree.h>

#include "bpipep.h"

/*

  PROCEDURE
  bpipe_output - create a binary pipe output channel

  DESCRIPTION
  @code{bpipe_output} attaches an output stream and opens it.

  The parameter @code{output} is a string describing the output file or
  device associated with the output channel.  If it is the string
  @samp{stdout}, the output channel is associated with the standard
  output stream.

  RETURNS
  It returns an output handle upon success, @code{NULL} upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

    @table @code
    @item BPEBADARG
          The passed path could not be opened.  Check @code{errno} for more
          information.

    @item BPEBADPIPE
          The @code{BPipe}'s header has already been mapped or written.  Output
          channels can only be opened before mapping or writing the header

    @item BPENOMEM
      a memory allocation failed
    @end table

*/
BPipeOutput *
bpipe_output(
  BPipe *bpipe,		/* a pointer to the binary pipe structure */
  const char *path	/* the path to the channel to attach to this output
                           descriptor */
)
{
  BPipeOutput *bpo = NULL;

  /* can't create a new output channel if the hdr has already been written */
  if ( bpipe->hdr_written || bpipe->hdr_mapped )
  {
    bpipe_errno = BPEBADPIPE;
    return NULL;
  }

  /* allocate Output Channel structure */
  if ( NULL == (bpo = (BPipeOutput *) malloc(sizeof(BPipeOutput))))
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  if ( NULL == (bpo->dpktf_dels = rbtree_new(rbtree_node_cmp_v)) )
  {
    free( bpo );
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  if ( path == NULL || path[0] == '\0' )
  {
    bpipe_errno = BPEBADARG;
    errno = ENOENT;
    return NULL;
  }

  /* allocate IOchannel */
  if ( NULL == (bpo->channel = bpipe_iochannel_new(path) ) )
  {
    return NULL;
  }

  if ( bpipe_iochannel_open( bpo->channel, "w" ) )
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  bpo->xmap = NULL;
  bpo->nmap = 0;
  bpo->matrix = NULL;
  bpo->buf = NULL;

  if ( NULL == ( bpo->matrix = sll_new( rbtree_node_cmp_v ) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  if ( NULL == ( bpo->dpktf = sll_new( SLL_NULL_CMP ) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  /* insert output channel into list of channels */
  if ( sll_insert_head( bpipe->output, bpo) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  return bpo;
}

/*

  PROCEDURE
  bpipe_output_delete - delete an output descriptor structure

  DESCRIPTION
  This routine frees the memory associated with an output descriptor structure
  and its associated data.  The I/O channel must already have been closed.

*/
void
bpipe_output_delete(
  BPipeOutput *bpo	/* the output descriptor structure to delete */
)
{

  if (bpo)
  {
    bpipe_iochannel_delete( bpo->channel );

    if ( bpo->dpktf_dels )
      rbtree_delete( bpo->dpktf_dels, RBTREE_NULL_DELETE, POST_ORDER,
		     LEFT_TO_RIGHT );

    if ( bpo->matrix )
      sll_delete( bpo->matrix, (SLL_DELETE) bpipe_matrix_override_delete );

    if ( bpo->dpktf )
      sll_delete( bpo->dpktf, free );

    if ( bpo->xmap )
      free( bpo->xmap );

    if ( bpo->buf )
      free( bpo->buf );

    free(bpo);
  }

}
