/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2005-2011 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include "bpipep.h"
#include <bpipe/datatypes.h>

static char* def_formats[] =
{
  "unknown type",	/*	BPDType_NOTYPE		*/
  "%s",			/*	BPDType_char  		*/
  "%.14g",		/*	BPDType_double		*/
  "%i",			/*	BPDType_int		*/
  "%u",			/*	BPDType_uint		*/
  "%.14g\t%.14g",		/*	BPDType_DVector2	*/
  "%.14g\t%.14g\t%.14g",		/*	BPDType_DVector3	*/
  "%i\t%i",		/*	BPDType_IVector2	*/
  "%i\t%i\t%i",		/*	BPDType_IVector3	*/
  "%u\t%u",		/*	BPDType_UIVector2	*/
  "%u\t%u\t%u",		/*	BPDType_UIVector3	*/
  "%.14g\t%.14g",		/*	BPDType_DComplex	*/
  "%.14g\t%.14g\t%.14g\t%.14g",	/*	BPDType_DCVector2	*/
  "%.14g\t%.14g\t%.14g\t%.14g\t%.14g\t%.14g"	/*	BPDType_DCVector3 */
};

/*

  PROCEDURE
  bpipe_sprintf - Format a @code{BPipe} data field for output.

  DESCRIPTION
  This routine will format a data field for output.  It fills a
  user-provided array with characters representing the data, in much
  the same way that @code{sprintf} does.  It uses default output
  formats for the various data types, which may be changed by
  providing a new array of output formats.

  RETURNS
  It returns the same value as does @code{sprintf}.

 */
int
bpipe_sprintf(
  char *	s,		/* a buffer to hold the formatted output. it
				   should be large enough to hold any output
				   from any data type */
  void *	data,		/* a pointer to the data to be formatted */
  BPDataType	type,		/* the data's @code{BPipe} datatype code */
  char *	formats[]	/* an array of pointers to formats,
				   with @code{BPDType_num} elements.
				   These should be in the order of the
				   @code{enum}'s (see @file{bpipe.h})
				   The first element (corresponding to
				   @code{BPDType_NOTYPE}, should
				   contain an error message.  If this
				   argument is @code{NULL}, the
				   default formats are used*/
)
{

  if ( NULL == formats )
    formats = def_formats;

  switch ( type )
  {
  case BPDType_char:
    return sprintf( s, formats[type], data );
    
  case BPDType_double:
    return sprintf( s, formats[type], *(double *) data );

  case BPDType_int:
    return sprintf( s, formats[type], *(int *) data );

  case BPDType_uint:
    return sprintf( s, formats[type], *(unsigned int *) data );

  case BPDType_DVector2:
    return sprintf( s, formats[type],
		    ((DVector2 *) data)->x,
		    ((DVector2 *) data)->y
		    );

  case BPDType_DVector3:
    return sprintf( s, formats[type],
		    ((DVector3 *) data)->x,
		    ((DVector3 *) data)->y,
		    ((DVector3 *) data)->z
		    );

  case BPDType_IVector2:
    return sprintf( s, formats[type],
		    ((IVector2 *) data)->x,
		    ((IVector2 *) data)->y
		    );

  case BPDType_IVector3:
    return sprintf( s, formats[type],
		    ((IVector3 *) data)->x,
		    ((IVector3 *) data)->y,
		    ((IVector3 *) data)->z
		    );

  case BPDType_UIVector2:
    return sprintf( s, formats[type],
		    ((UIVector2 *) data)->x,
		    ((UIVector2 *) data)->y
		    );

  case BPDType_UIVector3:
    return sprintf( s, formats[type],
		    ((UIVector3 *) data)->x,
		    ((UIVector3 *) data)->y,
		    ((UIVector3 *) data)->z
		    );

  case BPDType_DComplex:
    return sprintf( s, formats[type],
		    ((DComplex *) data)->r,
		    ((DComplex *) data)->i
		    );

  case BPDType_DCVector2:
    return sprintf( s, formats[type],
		    ((DCVector2 *) data)->q1.r,
		    ((DCVector2 *) data)->q1.i,
		    ((DCVector2 *) data)->q2.r,
		    ((DCVector2 *) data)->q2.i
		    );

  case BPDType_DCVector3:
    return sprintf( s, formats[type],
		    ((DCVector3 *) data)->q1.r,
		    ((DCVector3 *) data)->q1.i,
		    ((DCVector3 *) data)->q2.r,
		    ((DCVector3 *) data)->q2.i,
		    ((DCVector3 *) data)->q3.r,
		    ((DCVector3 *) data)->q3.i
		    );

  default:
  return sprintf( s, formats[0], data );
  }
    
  /* NOT REACHED */
}

